/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package es.rickyepoderi.compiler;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileReader;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.env.NameEnvironmentAnswer;

/**
 * The compiler environment in ECJ is just a way for returning classes that
 * are needed for compilation. This implementation just uses a class loader
 * to locate needed classes. The only exception is the class which is
 * being compiled, if the requested class is that one, the class loader is not
 * used.
 * 
 * @author ricky
 */
public class CompilerEnvironment implements INameEnvironment {

    /**
     * The class name being compiled.
     */
    private String className = null;
    
    /**
     * The source of the class being compiled.
     */
    private char[] source = null;
    
    /**
     * The class loader in which this class delegates.
     */
    private ClassLoader classLoader = null;
    
    /**
     * Constructor via arguments.
     * 
     * @param classLoader The class loader to use
     * @param className The main class name being compiled
     * @param source The source of the java file
     */
    public CompilerEnvironment(ClassLoader classLoader, String className, char[] source) {
        this.classLoader = classLoader;
        this.className = className;
        this.source = source;
    }
    
    /**
     * Original method of INameEnvironment interface based on char arrays. 
     * @param compoundTypeName The complete class name
     * @return The environment answer for that name
     */
    @Override
    public NameEnvironmentAnswer findType(char[][] compoundTypeName) {
        StringBuilder result = new StringBuilder();
        String sep = "";
        for (char[] compoundTypeName1 : compoundTypeName) {
            result.append(sep);
            result.append(new String(compoundTypeName1));
            sep = ".";
        }
        return findType(result.toString());
    }

    /**
     * Original method of INameEnvironment interface based on char arrays.
     * @param typeName The last part of the name
     * @param packageName The parent package name
     * @return The environment answer for that name
     */
    @Override
    public NameEnvironmentAnswer findType(char[] typeName, char[][] packageName) {
        StringBuilder result = new StringBuilder();
        String sep = "";
        for (char[] packageName1 : packageName) {
            result.append(sep);
            result.append(packageName1);
            sep = ".";
        }
        result.append(sep);
        result.append(typeName);
        return findType(result.toString());
    }
    
    /**
     * Return the type of a requested class name. If the class name passed is
     * the one currently being compiled a compiletion unit is returned (with
     * that name and source) else the class loader is used to return the type.
     * @param newClassName The class name requested
     * @return The environment answer for that name
     */
    private NameEnvironmentAnswer findType(String newClassName) {
        try {
            if (newClassName.equals(className)) {
                ICompilationUnit compilationUnit = new CompilationUnit(newClassName, source);
                return new NameEnvironmentAnswer(compilationUnit, null);
            }
            String resourceName = newClassName.replace('.', '/') + ".class";
            try (InputStream is = classLoader.getResourceAsStream(resourceName)) {
                if (is != null) {
                    byte[] classBytes;
                    byte[] buf = new byte[8192];
                    ByteArrayOutputStream baos = new ByteArrayOutputStream(buf.length);
                    int count;
                    while ((count = is.read(buf, 0, buf.length)) > 0) {
                        baos.write(buf, 0, count);
                    }
                    baos.flush();
                    classBytes = baos.toByteArray();
                    char[] fileName = newClassName.toCharArray();
                    ClassFileReader classFileReader = new ClassFileReader(classBytes, fileName, true);
                    return new NameEnvironmentAnswer(classFileReader, null);
                }
            }
        } catch (IOException | org.eclipse.jdt.internal.compiler.classfmt.ClassFormatException e) {
            throw new IllegalStateException(e);
        }
        return null;
    }

    /**
     * Return if the passed is a package (or a class).
     * @param result The name of the package to check
     * @return true if the passed name is a package false otherwise
     */
    private boolean isPackage(String result) {
        if (result.equals(className)) {
            // the compiled class is not a  package
            return false;
        }
        String resourceName = result.replace('.', '/') + ".class";
        try (InputStream is = classLoader.getResourceAsStream(resourceName)) {
            return is == null;
        } catch(IOException e) {
            throw new IllegalStateException(e);
        }
    }

    /**
     * Original method of INameEnvironment interface based on char arrays.
     * @param parentPackageName The parent package name
     * @param packageName The name of current package
     * @return true if the passed name is a package false otherwise
     */
    @Override
    public boolean isPackage(char[][] parentPackageName,
            char[] packageName) {
        StringBuilder result = new StringBuilder();
        String sep = "";
        if (parentPackageName != null) {
            for (char[] parentPackageName1 : parentPackageName) {
                result.append(sep);
                String str = new String(parentPackageName1);
                result.append(str);
                sep = ".";
            }
        }
        String str = new String(packageName);
        if (Character.isUpperCase(str.charAt(0))) {
            if (!isPackage(result.toString())) {
                return false;
            }
        }
        result.append(sep);
        result.append(str);
        return isPackage(result.toString());
    }

    @Override
    public void cleanup() {
        // NOOP
    }
}
