/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package es.rickyepoderi.compiler;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import org.apache.jasper.Options;
import org.eclipse.jdt.internal.compiler.Compiler;
import org.eclipse.jdt.internal.compiler.DefaultErrorHandlingPolicies;
import org.eclipse.jdt.internal.compiler.IErrorHandlingPolicy;
import org.eclipse.jdt.internal.compiler.IProblemFactory;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.problem.DefaultProblemFactory;

/**
 * A simple compiler that using ECJ (Eclipse Compiler for Java) compiles a
 * source java file and loads the resulting classes into a custom class
 * loader and finally return a Class object for the main class in the file.
 * The Class can be used (using reflection) to create a new object instance,
 * execute some static method or whatever.
 * 
 * @author ricky
 */
public class SimpleCompiler {
    
    /**
     * The class loader to use in order to find classes for the compilation.
     */
    private ClassLoader classLoader = null;
    
    /**
     * Internal options used for compilation.
     */
    private Options opts = null;
    
    /**
     * Constructor using the class loader and the options to use for compiling.
     * @param classLoader The class loader used in the compilation
     * @param opts The options used for the compilation
     */
    public SimpleCompiler(ClassLoader classLoader, Options opts) {
        this.classLoader = classLoader;
        this.opts = opts;
    }
    
    /**
     * Method to compile a class, the source is passed as a char array.
     * @param className The class name to retrieve
     * @param source The source of the java file
     * @return The Class loaded by a custom class loader
     */
    public Class compile(String className, char[] source) {
        // environment
        INameEnvironment env = new CompilerEnvironment(classLoader, className, source);
        // policy
        IErrorHandlingPolicy policy = DefaultErrorHandlingPolicies.proceedWithAllProblems();
        // options
        Map<String,String> settings = new HashMap<>();
        settings.put(CompilerOptions.OPTION_LineNumberAttribute, CompilerOptions.GENERATE);
        settings.put(CompilerOptions.OPTION_SourceFileAttribute, CompilerOptions.GENERATE);
        settings.put(CompilerOptions.OPTION_ReportDeprecation, CompilerOptions.IGNORE);
        settings.put(CompilerOptions.OPTION_Encoding, opts.getJavaEncoding());
        if (opts.getClassDebugInfo()) {
            settings.put(CompilerOptions.OPTION_LocalVariableAttribute, CompilerOptions.GENERATE);
        }
        settings.put(CompilerOptions.OPTION_Source, opts.getCompilerSourceVM());
        settings.put(CompilerOptions.OPTION_TargetPlatform, opts.getCompilerTargetVM());
        settings.put(CompilerOptions.OPTION_Compliance, opts.getCompilerTargetVM());
        CompilerOptions options = new CompilerOptions(settings);
        // requestor
        CompilerRequestor requestor = new CompilerRequestor(className, classLoader);
        // problem factory
        IProblemFactory problemFactory = new DefaultProblemFactory(Locale.getDefault());
        String[] classNames = new String[] {className};
        ICompilationUnit[] compilationUnits = new ICompilationUnit[classNames.length];
        for (int i = 0; i < compilationUnits.length; i++) {
            compilationUnits[i] = new CompilationUnit(classNames[i], source);
        }
        // compile
        Compiler compiler = new Compiler(env, policy, options, requestor, problemFactory);
        compiler.compile(compilationUnits);
        if (requestor.isSuccess()) {
            return requestor.getResultClass();
        } else {
            throw new IllegalStateException("Compiling errors: " + requestor.getErrors());
        }
    }
    
    /**
     * Similar method but the class source is passed as a String.
     * @param className The class name to retrieve
     * @param source The source of the java file
     * @return The Class loaded by a custom class loader
     */
    public Class compile(String className, String source) {
        return compile(className, source.toCharArray());
    }
    
    /**
     * Similar method but using a reader for the source.
     * @param className The class name to retrieve
     * @param reader The source of the java file
     * @return The Class loaded by a custom class loader
     * @throws IOException Some error reading in the reader
     */
    public Class compile(String className, Reader reader) throws IOException {
        StringBuilder sb = new StringBuilder();
        int data = reader.read();
        while (data != -1) {
            sb.append((char) data);
            data = reader.read();
        }
        return compile(className, sb.toString());
    }
    
    /**
     * Test method.
     * @param args
     * @throws Exception 
     */
    static public void main(String[] args) throws Exception {
        String source
                = "package example;"
                + "import java.util.*;\n"
                + "public class HelloWorld {\n"
                + "  static public class Internal {\n"
                + "    public String sayHello() {\n"
                + "      return \"Hello \" + \"World!!!\";\n"
                + "    }\n"
                + "  }\n"
                + "  public static void sayHello() {\n"
                + "    System.out.println(new Internal().sayHello());\n"
                + "  }\n"
                + "}\n";
        
        String source2
                =  "package example;"
                + "import java.util.*;\n"
                + "public class HelloWorld {\n"
                + "  static public class Internal {\n"
                + "    public Object sayHello() {\n"
                + "      return \"Hello!\";"
                + "    }\n"
                + "  }\n"
                + "  public static void sayHello() {\n"
                + "    System.out.println(new Internal().sayHello());\n"
                + "  }\n"
                + "}\n";

        SimpleCompiler compiler = new SimpleCompiler(Thread.currentThread().getContextClassLoader(), new DefaultCompilerOptions());
        Class clazz = compiler.compile("example.HelloWorld", source2);
        System.err.println(clazz);
        Method method = clazz.getMethod("sayHello");
        method.invoke(null);
        javax.servlet.jsp.JspFactory.setDefaultFactory(new org.apache.jasper.runtime.JspFactoryImpl());
        clazz = compiler.compile("org.apache.jsp.sample2_jsp", new InputStreamReader(new FileInputStream("/home/ricky/sample2_jsp.java")));
        method = clazz.getMethod("test");
        method.invoke(null);
    }
}
